@@ -4,6 +4,6 @@ from . firefox import *
 
 __all__=[
 'Firefox', 'BY', 'KEYS',
-'JDict', 'TimeStamp', 'Moment', 'GSheet', 
+'JDict', 'TimeStamp', 'Moment', 'GSheet', 'GCal',
 'fire', 'requests', 'sqlite3', 'json', 'io', 'os', 'sys', 'platform', 'string', 'random', 're', 'datetime', 'tzlocal', 'pd', 'np', 'urlparse', 'BeautifulSoup', 
 ]
